'use strict';
/*
 *  Copyright (C) 1998-2024 by Northwoods Software Corporation. All Rights Reserved.
 */

import * as go from 'gojs';

/**
 * A way of reducing the number of cases where link segments of orthogonal links overlap.
 *
 * Typical setup:
 * ```
 *   myDiagram.routers.add(new AvoidsLinksRouter());
 * ```
 *
 * If you want to experiment with this extension, try the <a href="../../samples/AvoidsLinksRouter.html">AvoidsLinksRouter</a> sample.
 * @category Layout Extension
 */
export class AvoidsLinksRouter extends go.Router {
  private _linkSpacing: number;
  private _allsegs: go.List<go.List<_SegInfo>>;
  private _gridlines: go.List<go.List<_SegInfo>>;
  private _segs: go.List<_SegInfo>;
  private _avoidsNodes: boolean;
  private _iterations: number;

  constructor(init?: Partial<AvoidsLinksRouter>) {
    super();
    this.name = 'AvoidsLinksRouter';
    this._linkSpacing = 4;
    this._allsegs = new go.List();
    this._gridlines = new go.List();
    this._segs = new go.List();
    this._avoidsNodes = true; // whether to aggressively reduce link spacing to avoid node overlaps
    this._iterations = 1; // the Router should produce a good result without iterating in most cases, unless there are many coincident segments very close to Nodes
    if (init) Object.assign(this, init);
  }

  /**
   * Gets or sets the desired distance between Links that are separated by AvoidsLinksRouter.
   * If {@link avoidsNodes} is true, this value is the maximum allowed distance between such links,
   * but the distance could be smaller to avoid overlapping Nodes.
   *
   * The default value is 4.
   */
  get linkSpacing(): number {
    return this._linkSpacing;
  }
  set linkSpacing(value: number) {
    this._linkSpacing = value;
  }

  /**
   * Gets or sets whether the AvoidsLinksRouter should reduce spacing between separated Links to avoid overlap
   * with nearby Nodes. If {@link avoidsNodes} is false, all separated links will have a distance of exactly {@link linkSpacing},
   * even if this causes them to cross over a nearby Node.
   *
   * The default value is true.
   */
  get avoidsNodes(): boolean {
    return this._avoidsNodes;
  }
  set avoidsNodes(value: boolean) {
    this._avoidsNodes = value;
  }

  /**
   * Gets or sets the number of times the AvoidsLinksRouter will run iteratively.
   * In the vast majority of cases, one iteration should be enough and this parameter should not need to be modified.
   * Multiple iterations may be useful when numerous links coincide, particularly when they are close to Nodes with {@link avoidsNodes} set to true.
   *
   * The default value is 1.
   */
  get iterations(): number {
    return this._iterations;
  }
  set iterations(value: number) {
    this._iterations = value;
  }

  /**
   * Only route once, at the top level
   * @param {go.Group | go.Diagram} coll
   * @returns
   */
  override canRoute(coll: go.Group | go.Diagram) {
    if (!super.canRoute(coll)) return false;
    return true;
  }

  /**
   * Because of `canRoute`, the first argument here is always a Diagram.
   * @param {*} coll A Diagram or a Group
   * @param {*} diagram
   * @returns
   */
  override routeLinks(links: go.Set<go.Link>, coll: go.Diagram | go.Group) {
    const diagram = this.diagram;
    if (diagram === null) return;
    this.collectSegments(links, coll);
    let positions = null;
    if (this._avoidsNodes) {
      positions = diagram.getPositions(true, null, null);
    }
    this.adjustOverlaps(positions);

    // free segments used in this calculation
    for (const line of this._allsegs) {
      for (const seg of line) this._freeSegInfo(seg);
    }
    for (const line of this._gridlines) {
      for (const seg of line) this._freeSegInfo(seg);
    }
  }

  _allocSegInfo(): _SegInfo {
    const si = this._segs.pop();
    if (si) return si;
    return new _SegInfo();
  }

  _freeSegInfo(si: _SegInfo) {
    si.link = null;
    this._segs.push(si);
  }

  _coord(si: _SegInfo) {
    return si.vertical ? si.link!.getPoint(si.index).x : si.link!.getPoint(si.index).y;
  }

  _columnStart(si: _SegInfo) {
    return si.vertical ? si.link!.getPoint(si.index).y : si.link!.getPoint(si.index).x;
  }

  _columnEnd(si: _SegInfo) {
    return si.vertical ? si.link!.getPoint(si.index + 1).y : si.link!.getPoint(si.index + 1).x;
  }

  _column(si: _SegInfo, index: number, vertical: number) {
    return vertical ? si.link!.getPoint(index).y : si.link!.getPoint(index).x;
  }

  shouldRoute(link: go.Link, index: number, vertical: boolean) {
    if(
      this.isApprox(link.getPoint(index).x, link.getPoint(index + 1).x) &&
      this.isApprox(link.getPoint(index).y, link.getPoint(index + 1).y)) {
      return false;
    }
    return vertical ? this.shouldRouteV(link, index) : this.shouldRouteH(link, index);
  }

  shouldRouteH(link: go.Link, index: number) {
    let j = index;
    while (
      j > 0 &&
      this.isApprox(link.getPoint(j).y, link.getPoint(j - 1).y) &&
      !this.isApprox(link.getPoint(j).x, link.getPoint(j - 1).x)
    ) {
      j--;
    }
    if (j === 0) return false;
    j = index + 1;
    while (
      j < link.pointsCount - 1 &&
      this.isApprox(link.getPoint(j).y, link.getPoint(j + 1).y) &&
      !this.isApprox(link.getPoint(j).x, link.getPoint(j + 1).x)
    ) {
      j++;
    }
    if (j === link.pointsCount - 1) return false;
    return true;
  }

  shouldRouteV(link: go.Link, index: number) {
    let j = index;
    while (
      j > 0 &&
      !this.isApprox(link.getPoint(j).y, link.getPoint(j - 1).y) &&
      this.isApprox(link.getPoint(j).x, link.getPoint(j - 1).x)
    ) {
      j--;
    }
    if (j === 0) return false;
    j = index + 1;
    while (
      j < link.pointsCount - 1 &&
      !this.isApprox(link.getPoint(j).y, link.getPoint(j + 1).y) &&
      this.isApprox(link.getPoint(j).x, link.getPoint(j + 1).x)
    ) {
      j++;
    }
    if (j === link.pointsCount - 1) return false;
    return true;
  }

  collectSegments(links: go.Set<go.Link>, coll?: go.Diagram | go.Group) {
    this._allsegs.clear();
    this._gridlines.clear();
    const isDiagram = coll instanceof go.Diagram;
    let p: go.Point, q: go.Point;
    let found: boolean;
    let currentseg: _SegInfo = this._allocSegInfo();

    // add segments of links in the "invalid" links Set
    for (const l of links) {
      if (!((isDiagram && l.containingGroup === null) || l.containingGroup === coll)) continue;
      if (!l.isOrthogonal) continue;
      for (let i = 1; i < l.pointsCount - 2; i++) {
        p = l.getPoint(i);
        q = l.getPoint(i + 1);
        const vertical = this.isApprox(p.x, q.x) && !this.isApprox(p.y, q.y);

        if (!this.shouldRoute(l, i, vertical)) continue;

        const seginfo = this._allocSegInfo();
        seginfo.layer = NaN;
        seginfo.index = i;
        seginfo.link = l;
        seginfo.vertical = vertical;
        seginfo.invalid = true;
        seginfo._computeGeo();

        found = false;
        for(const line of this._allsegs) {
          if(this.isApprox(this._coord(line.first()!), this._coord(seginfo)) && line.first()!.vertical === vertical) {
            found = true;
            line.add(seginfo);
            break;
          }
        }
        if(!found) this._allsegs.add(new go.List([seginfo]));
      }
    }

    // split sets of segments into those which actually overlap
    for (const line of this._allsegs) {
      while(line.count > 0) {
        found = false;
        for (const seg of line) {
          if(seg.invalid) currentseg = seg;
          found = true;
          break;
        }

        if(!found) break;

        line.remove(currentseg);
        const newline = new go.List([currentseg]);
        found = true;
        while(found) {
          found = false;
          for (const otherseg of newline) {
            for (const seginfo of line) {
              const minI = Math.min(this._columnStart(seginfo), this._columnEnd(seginfo));
              const maxI = Math.max(this._columnStart(seginfo), this._columnEnd(seginfo));
              const minJ = Math.min(this._columnStart(otherseg), this._columnEnd(otherseg));
              const maxJ = Math.max(this._columnStart(otherseg), this._columnEnd(otherseg));
              if (minJ <= maxI && minI <= maxJ && seginfo.link !== otherseg.link) {
                line.remove(seginfo);
                newline.push(seginfo);
                found = true;
                break;
              }
            }
          }
        }
        this._gridlines.add(newline);
      }
    }
  }

  adjustOverlaps(positions: any) {
    const gridlines = this._gridlines;
    let count = 0;
    for(const line of gridlines) {
      if(line.size < 2) continue;
      const gridline = line.toArray();
      this.sortGridline(gridline);
      const vertical = gridline[0].vertical;

      // assign layers to overlapping segments greedily
      let maxlayer = 0;
      for (let i = 0; i < gridline.length; i++) {
        let newlayer = 0;
        const segI = gridline[i];
        for (let j = 0; j < i; j++) {
          const segJ = gridline[j];
          if(newlayer === segJ.layer) {
            newlayer++;
            count++;
          }
        }
        segI.layer = newlayer;
        maxlayer = Math.max(maxlayer, newlayer);
      }

      let realSpacing = this._linkSpacing;

      if (this._avoidsNodes && maxlayer > 0) {
        // find the initial bounding box of the newly routed segments in the gridline
        let minColumn = Math.min(this._columnStart(gridline[0]), this._columnEnd(gridline[0])); // x if horizontal
        let maxColumn = Math.max(this._columnStart(gridline[0]), this._columnEnd(gridline[0]));
        const minCoord = this._coord(gridline[0]) - (maxlayer * this._linkSpacing) / 2; // x if vertical
        const maxCoord = this._coord(gridline[0]) + (maxlayer * this._linkSpacing) / 2;
        for (let i = 1; i < gridline.length; i++) {
          const seg = gridline[i];
          minColumn = Math.min(minColumn, Math.min(this._columnStart(seg), this._columnEnd(seg)));
          maxColumn = Math.max(maxColumn, Math.max(this._columnStart(seg), this._columnEnd(seg)));
        }

        // if there are node overlaps, find the minimum spacing that will avoid node overlaps with some padding
        if(vertical) {
          if (!positions.isUnoccupied(minCoord, minColumn, maxCoord - minCoord, maxColumn - minColumn)) {
            const availSpace = positions.maxUnoccupiedV(
              minColumn,
              maxColumn,
              this._coord(gridline[0]),
              maxCoord - minCoord
            );
            realSpacing = Math.min(this._linkSpacing, (2 * availSpace) / (1 + maxlayer));
          }
        } else {
          if (!positions.isUnoccupied(minColumn, minCoord, maxColumn - minColumn, maxCoord - minCoord)) {
            const availSpace = positions.maxUnoccupiedH(
              minColumn,
              maxColumn,
              this._coord(gridline[0]),
              maxCoord - minCoord
            );
            realSpacing = Math.min(this._linkSpacing, (2 * availSpace) / (1 + maxlayer));
          }
        }
      }

      // now that segment layers have been found, reroute segments according to their layer
      for (let i = 0; i < gridline.length; i++) {
        const seg = gridline[i];
        if (seg.link === null) continue;
        const newcoord = this._coord(seg) + (seg.layer - maxlayer / 2) * realSpacing;
        seg.link.startRoute();
        if(vertical) {
          seg.link.setPoint(seg.index, new go.Point(newcoord, this._columnStart(seg)));
          seg.link.setPoint(seg.index + 1, new go.Point(newcoord, this._columnEnd(seg)));
        } else {
          seg.link.setPoint(seg.index, new go.Point(this._columnStart(seg), newcoord));
          seg.link.setPoint(seg.index + 1, new go.Point(this._columnEnd(seg), newcoord));
        }
        seg.link.commitRoute();
      }
    }
  }

  partialSort(arr: Array<any>, start: number, end: number, f: (a: any, b: any) => number) {
    const preSorted = arr.slice(0, start);
    const postSorted = arr.slice(end);
    const sorted = arr.slice(start, end).sort(f);
    arr.length = 0;
    // eslint-disable-next-line prefer-spread
    arr.push.apply(arr, preSorted.concat(sorted).concat(postSorted));
    return arr;
  }

  endpointComparer(seg1: _SegInfo, seg2: _SegInfo, idx1: number, idx2: number) {
    const i1 = idx1;
    const i2 = idx2;
    if (seg1.link === null) return 0;
    if (seg2.link === null) return 0;
    let vertical = seg1.vertical ? 1 : 0;

    let min1 = Math.min(this._column(seg1, idx1, vertical), this._column(seg1, idx1 + 1, vertical));
    let max1 = Math.max(this._column(seg1, idx1, vertical), this._column(seg1, idx1 + 1, vertical));
    let min2 = Math.min(this._column(seg2, idx2, vertical), this._column(seg2, idx2 + 1, vertical));
    let max2 = Math.max(this._column(seg2, idx2, vertical), this._column(seg2, idx2 + 1, vertical));

    let result = 0;
    if (this.isApprox(max2, max1)) result = min1 - min2;
    else result = max2 - max1;

    while (result === 0) {
      if (idx1 > 0 && idx2 > 0) {
        idx1--;
        idx2--;
        vertical = 1 - vertical;
      } else break;

      min1 = Math.min(this._column(seg1, idx1, vertical), this._column(seg1, idx1 + 1, vertical));
      max1 = Math.max(this._column(seg1, idx1, vertical), this._column(seg1, idx1 + 1, vertical));
      min2 = Math.min(this._column(seg2, idx2, vertical), this._column(seg2, idx2 + 1, vertical));
      max2 = Math.max(this._column(seg2, idx2, vertical), this._column(seg2, idx2 + 1, vertical));

      if (this.isApprox(max2, max1)) result = min1 - min2;
      else result = max2 - max1;
    }

    idx1 = i1;
    idx2 = i2;
    vertical = seg1.vertical ? 1 : 0;

    while (result === 0) {
      if (idx1 < seg1.link.pointsCount - 2 && idx2 < seg2.link.pointsCount - 2) {
        idx1++;
        idx2++;
        vertical = 1 - vertical;
      } else break;

      min1 = Math.min(this._column(seg1, idx1, vertical), this._column(seg1, idx1 + 1, vertical));
      max1 = Math.max(this._column(seg1, idx1, vertical), this._column(seg1, idx1 + 1, vertical));
      min2 = Math.min(this._column(seg2, idx2, vertical), this._column(seg2, idx2 + 1, vertical));
      max2 = Math.max(this._column(seg2, idx2, vertical), this._column(seg2, idx2 + 1, vertical));

      if (this.isApprox(max2, max1)) result = min1 - min2;
      else result = max2 - max1;
    }

    return seg1.geo % 2 === 0 ? result : -result;
  }

  sortGridline(gridline: any) {
    // first sort all segments according to their seginfo.geo value, so that segments with the same geometry can be sorted against each other
    gridline.sort((seg1: _SegInfo, seg2: _SegInfo) => seg2.geo - seg1.geo);

    // find number of segments with each geometry
    let numGeo0 = 0;
    let numGeo1 = 0;
    let numGeo2 = 0;
    let numGeo3 = 0;

    for (let i = 0; i < gridline.length; i++) {
      switch (gridline[i].geo) {
        case 0:
          numGeo0++;
          break;
        case 1:
          numGeo1++;
          break;
        case 2:
          numGeo2++;
          break;
        case 3:
          numGeo3++;
          break;
      }
    }

    const n1 = numGeo0;
    const n2 = numGeo0 + numGeo1;
    const n3 = numGeo0 + numGeo1 + numGeo2;
    const n4 = numGeo0 + numGeo1 + numGeo2 + numGeo3;

    // sort segments with geo = 0
    if (numGeo0 > 1) {
      this.partialSort(gridline, 0, n1, (seg1, seg2) =>
        this.endpointComparer(seg1, seg2, seg1.index, seg2.index)
      );
    }
    // sort segments with geo = 1
    if (numGeo1 > 1) {
      this.partialSort(gridline, n1, n2, (seg1, seg2) =>
        this.endpointComparer(seg1, seg2, seg1.index, seg2.index)
      );
    }
    // sort segments with geo = 2
    if (numGeo2 > 1) {
      this.partialSort(gridline, n2, n3, (seg1, seg2) =>
        this.endpointComparer(seg1, seg2, seg1.index, seg2.index)
      );
    }
    // sort segments with geo = 3
    if (numGeo3 > 1) {
      this.partialSort(gridline, n3, n4, (seg1, seg2) =>
        this.endpointComparer(seg1, seg2, seg1.index, seg2.index)
      );
    }
  }

  isApprox(a: number, b: number) {
    const d = a - b;
    return d > -0.5 && d < 0.5;
  }
}

// Internal information about each segment
class _SegInfo {
  vertical: boolean;
  index: number;
  layer: number;
  link: go.Link | null;
  geo: number;
  invalid: boolean;
  assigned: boolean;
  constructor() {
    this.vertical = false; // true if the segment is vertical, false if horizontal
    this.index = NaN;
    this.layer = NaN;
    this.link = null;
    this.geo = 0; // geometry of the segment. horizontally, 0 = down->over->down, 1 = up->over->up, 2 = up->over->down, 3 = down->over->up
    this.invalid = false;
    this.assigned = false;
  }

  // generic compute this.geo
  _computeGeo() {
    if(this.vertical) this._computeGeoV();
    else this._computeGeoH();
  }

  // compute this.geo if horizontal
  _computeGeoH() {
    if (this.link === null) return;
    // j1 and j2 are the indices of the next orthogonal bends (not necessarily the next point if there are redundant link points)
    let j1 = this.index - 1;
    let j2 = this.index + 2;
    while (j1 > 0 && Math.abs(this.link.getPoint(j1).x - this.link.getPoint(j1 - 1).x) < 0.5) j1--;
    while (
      j2 < this.link.pointsCount - 1 &&
      Math.abs(this.link.getPoint(j2).x - this.link.getPoint(j2 + 1).x) < 0.5
    ) {
      j2++;
    }

    const y1 = this.link.getPoint(j1).y;
    const y2 = this.link.getPoint(j2).y;
    const coord = this.link.getPoint(this.index).y;
    const columnStart = this.link.getPoint(this.index).x;
    const columnEnd = this.link.getPoint(this.index + 1).x;

    if (columnStart < columnEnd) {
      if (y1 < coord && y2 > coord) this.geo = 0;
      else if (y1 > coord && y2 < coord) this.geo = 1;
      else if (y1 > coord && y2 > coord) this.geo = 2;
      else if (y1 < coord && y2 < coord) this.geo = 3;
    } else {
      if (y2 < coord && y1 > coord) this.geo = 0;
      else if (y2 > coord && y1 < coord) this.geo = 1;
      else if (y2 > coord && y1 > coord) this.geo = 2;
      else if (y2 < coord && y1 < coord) this.geo = 3;
    }
  }

  // compute this.geo if vertical
  _computeGeoV() {
    if (this.link === null) return;
    // j1 and j2 are the indices of the next orthogonal bends (not necessarily the next point if there are redundant link points)
    let j1 = this.index - 1;
    let j2 = this.index + 2;
    while (j1 > 0 && Math.abs(this.link.getPoint(j1).y - this.link.getPoint(j1 - 1).y) < 0.5) j1--;
    while (
      j2 < this.link.pointsCount - 1 &&
      Math.abs(this.link.getPoint(j2).y - this.link.getPoint(j2 + 1).y) < 0.5
    ) {
      j2++;
    }
    const x1 = this.link.getPoint(j1).x;
    const x2 = this.link.getPoint(j2).x;
    const coord = this.link.getPoint(this.index).x;
    const columnStart = this.link.getPoint(this.index).y;
    const columnEnd = this.link.getPoint(this.index + 1).y;

    if (columnStart < columnEnd) {
      if (x1 < coord && x2 > coord) this.geo = 0;
      else if (x1 > coord && x2 < coord) this.geo = 1;
      else if (x1 > coord && x2 > coord) this.geo = 2;
      else if (x1 < coord && x2 < coord) this.geo = 3;
    } else {
      if (x2 < coord && x1 > coord) this.geo = 0;
      else if (x2 > coord && x1 < coord) this.geo = 1;
      else if (x2 > coord && x1 > coord) this.geo = 2;
      else if (x2 < coord && x1 < coord) this.geo = 3;
    }
  }
} // end of _SegInfo
