/*
 *  Copyright (C) 1998-2024 by Northwoods Software Corporation. All Rights Reserved.
 */

/*
 * This is an extension and not part of the main GoJS library.
 * Note that the API for this class may change with any version, even point releases.
 * If you intend to use an extension in production, you should copy the code to your own source directory.
 * Extensions can be found in the GoJS kit under the extensions or extensionsJSM folders.
 * See the Extensions intro page (https://gojs.net/latest/intro/extensions.html) for more information.
 */

import * as go from 'gojs';

/**
 * A class for automatically adjusting link label placement on a diagram.
 *
 * By utilising greedy heuristics to make a direct-descent algorithm, labels are moved to a more
 * appropriate position by considering any overlap with labels and/or nodes and the label's distance
 * to its link. If better positions are found, the labels are moved.
 *
 * The constructor takes four arguments:
 *   - `precision` ***number*** an optional argument for speed vs. precision (1-5). 1: slow, precise ; 5 - fast, inexact
 *   - `detectGroups` ***boolean*** an optional argument to acknowledge group overlap in positioning
 *
 * If you want to experiment with this extension, try the <a href="../../samples/LinkLabelRouter.html">LinkLabelRouter</a> sample.
 * @category Layout Extension
 */
export class LinkLabelRouter extends go.Router {
  private _precision: number;
  private _detectGroups: boolean;
  private _nearby: go.Set<go.GraphObject>;
  private _overheadAllLabels: go.Map<go.GraphObject, go.Spot | null> | null;
  private _selectedGraphObj: any;

  /**
   * Constructs a LinkLabelRouter for a given {@link Diagram} or {@link GraphObject} based on the
   * optional parameters.
   *
   * Also sets up event listeners to run the algorithm when an update is made to the Diagram
   *
   * @param {Partial<LinkLabelRouter>=} init an Object with optional properties to be set on the new LinkLabelRouter
   */
  constructor(init?: Partial<LinkLabelRouter>) {
    super();
    this.name = 'LinkLabel';
    this._precision = 3;
    this._detectGroups = false;
    // Collect nearby labels and/or nodes
    this._nearby = new go.Set();
    // Collect all labels and their original alignmentFocus values
    // This is populated the first time routing is run
    this._overheadAllLabels = null;
    if (init) Object.assign(this, init);
  }

  /**
   * Gets or sets the {@link GraphObject} associated with LinkLabelRouter
   */
  get selectedGraphObj() {
    return this._selectedGraphObj;
  }
  set selectedGraphObj(val) {
    if (val !== this._selectedGraphObj) {
      this._selectedGraphObj = val;
    }
  }

  /**
   * Gets or sets the precision value associated with LinkLabelRouter.
   *
   * The value ranges between 1-5 with 1 having the highest precision but slower speed
   * and 5 having the lowest precision but faster speed. Setting 5 will move labels a maximum
   * of one time.
   *
   * The default value is 3. If the user inputs a number not between 1-5, the default is returned.
   */
  get precision() {
    return this._precision;
  }
  set precision(val) {
    if (typeof val !== 'number') {
      throw new Error('new precision property value is not a number: ' + val);
    }
    if (val < 1) val = 1;
    else if (val > 5) val = 5;
    if (val !== this._precision) {
      this._precision = val;
    }
  }

  /**
   * Gets or sets the detectGroups boolean associated with LinkLabelRouter
   *
   * Overlap between a label and a group may or may not be desirable. If detectGroup === true, the class
   * will still ignore groups that completely encase the link i.e. the to- and from-nodes are both in
   * the group.
   *
   * IMPORTANT: computesBoundsIncludingLinks MUST be set as false on the GroupTemplate
   *
   * The default value is false.
   */
  get detectGroups() {
    return this._detectGroups;
  }
  set detectGroups(val) {
    if (typeof val !== 'boolean') {
      throw new Error('new detectGroups property value is not a boolean: ' + val);
    }
    if (val !== this._detectGroups) {
      this._detectGroups = val;
    }
  }

  /**
   * Collects all labels and original alignmentFocus values
   *
   * KEY: label
   * VALUE: alignmentFocus
   *
   * @returns {go.Map<go.GraphObject, go.Spot | null>}
   */
  private _collectAllLabelsDiagram(diagram: go.Diagram) {
    const labelObj = new go.Map<go.GraphObject, go.Spot | null>();
    diagram.links.each((l) => {
      l.elements.each((e) => {
        if (this.isLabel(e)) {
          labelObj.set(e, e.alignmentFocus);
        }
      });
    });
    return labelObj;
  }

  /**
   * Runs the link label adjust algorithm
   *
   * @returns {void}
   */
  override routeLinks(links: go.Set<go.Link>, coll: go.Diagram | go.Group) {
    const diagram = this.diagram;
    if (diagram === null) return;
    //??? This code needs to consider the LINKS and COLL arguments correctly
    // const container = coll instanceof go.Diagram ? null : coll;
    // links.each(link => {
    //   if (coll && link.containingGroup !== container) return;
    //   ...
    // });
    if (this._overheadAllLabels === null) {
      this._overheadAllLabels = this._collectAllLabelsDiagram(diagram);
    }
    let nearObj = null;
    if (!this.selectedGraphObj) {
      // If obj is undefined, collect all labels in the Diagram
      nearObj = this._overheadAllLabels;
    } else {
      // Collect labels associated (by location) with the particular GraphObject
      nearObj = this._collectAllLabelsNear(diagram);
    }
    const self = this;
    // Iterate through all collected labels
    nearObj.each(function (kvp) {
      const graphObj = kvp.key;
      const alignmentFocusOriginal = kvp.value;
      if (alignmentFocusOriginal === null) return;
      // Find the current position score
      self._nearby = self._findObjectsClose(graphObj, diagram);
      const currentScore = self._cost(graphObj);
      // Find the original focus values
      const originalOffX = alignmentFocusOriginal.offsetX;
      const originalOffY = alignmentFocusOriginal.offsetY;
      // Find the current focus values
      const currentOffX = graphObj.alignmentFocus.offsetX;
      const currentOffY = graphObj.alignmentFocus.offsetY;
      // Find the original position score
      self._changeFocusLabel(graphObj, originalOffX - currentOffX, originalOffY - currentOffY);
      const originalScore = self._cost(graphObj);
      self._changeFocusLabel(graphObj, currentOffX - originalOffX, currentOffY - originalOffY);
      // Calculate a vector movement from overlapping labels/nodes and distance to link
      const finalVector = self._vectorMovement(graphObj);
      self._changeFocusLabel(graphObj, -finalVector[0], -finalVector[1]);
      const movedScore = self._cost(graphObj);
      // Check to see the best position of the three
      const minScore = Math.min(originalScore, currentScore, movedScore);
      if (minScore === currentScore) {
        self._changeFocusLabel(graphObj, finalVector[0], finalVector[1]);
      } else if (minScore === originalScore) {
        self._changeFocusLabel(
          graphObj,
          finalVector[0] + originalOffX - currentOffX,
          finalVector[1] + originalOffY - currentOffY
        );
      }
    });
    // Precision movement following an inital vector movement
    if (this.precision !== 5) {
      nearObj.each(function (kvp) {
        const graphObj = kvp.key;
        self._nearby = self._findObjectsClose(graphObj, diagram);
        // Small precision movements to find local minima
        self._differentialCalculatorVector(graphObj);
      });
    }
    this.selectedGraphObj = undefined;
    return;
  }

  /**
   * Check to see if a GraphObject is a Link Label
   *
   * @hidden @internal
   * @param {GraphObject} obj
   * @returns {boolean}
   */
  isLabel(obj: go.GraphObject) {
    if (!obj) return false;
    const link = obj.panel;
    if (obj.panel === null) return false;
    if (link instanceof go.Link) {
      if (
        obj instanceof go.Shape &&
        (obj.isPanelMain ||
          obj.panel.findMainElement() === obj ||
          obj.fromArrow !== 'None' ||
          obj.toArrow !== 'None')
      ) {
        return false;
      } else {
        return true;
      }
    }
    return false;
  }

  /**
   * Collects all labels and corresponding alignmentFocus values
   * associated with a SelectionMoved GraphObject
   *
   * @hidden @internal
   * @param {GraphObject} obj
   * @returns {go.Map<go.GraphObject, go.Spot | null>}
   */
  private _collectAllLabelsNear(diagram: go.Diagram): go.Map<go.GraphObject, go.Spot | null> {
    const labelObj = new go.Map<go.GraphObject, go.Spot | null>();
    if (this.selectedGraphObj === undefined) return new go.Map();
    const self = this;
    // Node --> Links --> Labels
    if (this.selectedGraphObj instanceof go.Node) {
      const linksIterator = this.selectedGraphObj.findLinksConnected().iterator;
      while (linksIterator.next()) {
        const link = linksIterator.value;
        link.elements.each((e) => {
          if (self.isLabel(e)) {
            labelObj.set(e, self._overheadAllLabels ? self._overheadAllLabels.get(e) : null);
          }
        });
      }
      // Labels intersecting moved node
      const objRect = this.selectedGraphObj.getDocumentBounds();
      diagram.findObjectsIn(
        objRect,
        function (p) {
          return self.isLabel(p) ? p : null;
        },
        function (n) {
          labelObj.set(n, self._overheadAllLabels ? self._overheadAllLabels.get(n) : null);
          return true;
        },
        true
      );
      // Link --> Labels
    } else if (this.selectedGraphObj instanceof go.Link) {
      this.selectedGraphObj.elements.each((e) => {
        if (this.isLabel(e)) {
          labelObj.set(e, this._overheadAllLabels ? this._overheadAllLabels.get(e) : null);
        }
      });
      // Label --> Label
    } else if (this.isLabel(this.selectedGraphObj)) {
      labelObj.set(
        this.selectedGraphObj,
        this._overheadAllLabels ? this._overheadAllLabels.get(this.selectedGraphObj) : null
      );
    }
    return labelObj;
  }

  /**
   * Find labels and nodes close to the Label
   *
   * @hidden @internal
   * @param {GraphObject} label
   * @returns {go.Set<go.GraphObject>}
   */
  private _findObjectsClose(label: go.GraphObject, diagram: go.Diagram) {
    const rect = label.getDocumentBounds();
    const radius = 0.75 * Math.hypot(rect.width, rect.height);
    const objects = new go.Set<go.GraphObject>();
    const self = this;
    if (this.detectGroups === false) {
      diagram.findObjectsNear(
        rect.center,
        radius,
        function (p) {
          return !(p instanceof go.Group) && (self.isLabel(p) || p instanceof go.Node) ? p : null;
        },
        function (n) {
          objects.add(n);
          return true;
        },
        true
      );
      return objects;
    } else {
      diagram.findObjectsNear(
        rect.center,
        radius,
        function (p) {
          return self.isLabel(p) || p instanceof go.Node ? p : null;
        },
        function (n) {
          objects.add(n);
          return true;
        },
        true
      );
      return objects;
    }
  }

  /**
   * Finds the score of a label's current position from overlap and distance to link
   *
   * @hidden @internal
   * @param {GraphObject} label
   * @returns {number}
   */
  private _cost(label: go.GraphObject) {
    return this._distanceToLink(label) + this._overlap(label);
  }

  /**
   * An ambiguity score for the distance of a label to its link
   *
   * @hidden @internal
   * @param {GraphObject} label
   * @returns {number}
   */
  private _distanceToLink(label: go.GraphObject) {
    const link = label.panel;
    if (!(link instanceof go.Link)) return 0;
    if (link.path === null) return 0;
    // Calculate centre point of label
    const pointLabel = link.path.getLocalPoint(label.getDocumentPoint(go.Spot.Center));
    // Calculate fraction (0-1) down link for label
    const fractionLinkForLabel = link.geometry.getFractionForPoint(pointLabel);
    // Caluclate point on link for that fraction
    const pointLink = link.geometry.getPointAlongPath(fractionLinkForLabel);
    const size = this._sizeOfGraphObj(label);
    const distance = Math.pow(pointLabel.distanceSquaredPoint(pointLink), 0.5);
    // Calculate a distance score
    const longtinConstant = 0.01;
    const distanceScore = longtinConstant * Math.pow(2, 4 + (10 * distance) / (size + 1));
    return distanceScore;
  }

  /**
   * Calculates an overlap score between a label and intersecting GraphObjects
   *
   * @hidden @internal
   * @param {GraphObject} label
   * @returns {number}
   */
  private _overlap(label: go.GraphObject) {
    const labelRect = label.getDocumentBounds();
    let overlapNodeScore = 0;
    let overlapLabelScore = 0;
    const labelConstant = 5;
    const nodeConstant = 12;
    const ignoreGroup = this._commonGroup(label);
    // Find corners of labelRect
    const l1 = labelRect.left;
    const t1 = labelRect.top;
    const r1 = labelRect.right;
    const b1 = labelRect.bottom;
    // Iterate through all nearby labels/nodes
    this._nearby.each((obj) => {
      if (label !== obj) {
        if (obj === ignoreGroup) return;
        const objRect = obj.getDocumentBounds();
        // Find corners of objRect
        const l2 = objRect.left;
        const t2 = objRect.top;
        const r2 = objRect.right;
        const b2 = objRect.bottom;
        if (obj instanceof go.Node) {
          const score = this._calculateOverlap(l1, t1, r1, b1, l2, t2, r2, b2);
          overlapNodeScore += score;
          // Check if the node completely intersects the main label
          if (score === this._area(obj)) {
            overlapNodeScore += 1000;
            overlapNodeScore += this._antiGrav(labelRect, objRect);
          }
        } else {
          const score = this._calculateOverlap(l1, t1, r1, b1, l2, t2, r2, b2);
          overlapLabelScore += score;
          // Check if the label completely intersects the main label
          if (score === this._area(obj)) {
            overlapLabelScore += 1000;
            overlapLabelScore += this._antiGrav(labelRect, objRect);
          }
        }
      }
    });
    // Overlap with a node is weighted more
    return nodeConstant * Math.pow(overlapNodeScore, 1.25) + labelConstant * overlapLabelScore;
  }

  /**
   * Calculates the area of overlap between two GraphObjects
   *
   * @hidden @internal
   * @param {Rect} firstRect
   * @param {Rect} secondRect
   * @returns {number}
   */
  private _calculateOverlap(
    l1: number,
    t1: number,
    r1: number,
    b1: number,
    l2: number,
    t2: number,
    r2: number,
    b2: number
  ) {
    const overlap =
      Math.max(0, Math.min(r1, r2) - Math.max(l1, l2)) *
      Math.max(0, Math.min(b1, b2) - Math.max(t1, t2));
    return overlap;
  }

  /**
   * Find the area of a GraphObject
   *
   * @hidden @internal
   * @param {GraphObject} graphObj
   * @returns {number}
   */
  private _area(graphObj: go.GraphObject) {
    const width = graphObj.getDocumentBounds().width;
    const height = graphObj.getDocumentBounds().height;
    return width * height;
  }

  /**
   * Contributes a negative score to a label if it completely intersects a GraphObject.
   *
   * The field allows a label to move out of a bad local minima.
   *
   * @hidden @internal
   * @param {Rect} firstRect
   * @param {Rect} secondRect
   * @returns {number}
   */
  private _antiGrav(firstRect: go.Rect, secondRect: go.Rect) {
    // Calculate the internal distances from edge to edge of the rectangles
    const top = secondRect.top - firstRect.top;
    const bottom = firstRect.bottom - secondRect.bottom;
    const left = secondRect.left - firstRect.left;
    const right = firstRect.right - secondRect.right;
    // Find the minimum distance
    const min = Math.min(top, bottom, left, right);
    const score = -1 * Math.pow(1 + 10 / (1 + min), 1.5);
    return score;
  }

  /**
   * Calculate a single vector movement from overlaps and distance to link
   *
   * @hidden @internal
   * @param {GraphObject} label
   * @returns {number[]}
   */
  private _vectorMovement(label: go.GraphObject) {
    const labelRect = label.getDocumentBounds();
    const labelCenter = labelRect.center;
    let finalVectorX = 0;
    let finalVectorY = 0;
    let finalMagnitude = 1;
    const ignoreGroup = this._commonGroup(label);
    // Find corners of labelRect
    const l1 = labelRect.left;
    const t1 = labelRect.top;
    const r1 = labelRect.right;
    const b1 = labelRect.bottom;
    this._nearby.each((graphObj) => {
      if (label === graphObj) return;
      if (graphObj === ignoreGroup) return;
      let graphObjRect = graphObj.getDocumentBounds();
      const graphObjCenter = graphObjRect.center;
      // Find corners of graphObjRect
      const l2 = graphObjRect.left;
      const t2 = graphObjRect.top;
      const r2 = graphObjRect.right;
      const b2 = graphObjRect.bottom;
      // If the nearby object intersects, a large vector is calculated
      if (this._doIntersect(l1, t1, r1, b1, l2, t2, r2, b2)) {
        // Find points of intersection
        let topLeftIntersecting = false;
        let topRightIntersecting = false;
        let bottomLeftIntersecting = false;
        let bottomRightIntersecting = false;
        let numCornersIntersecting = 0;
        let topLeft = new go.Point(l2, t2);
        if (labelRect.containsPoint(topLeft)) {
          topLeftIntersecting = true;
          numCornersIntersecting += 1;
        }
        let topRight = new go.Point(r2, t2);
        if (labelRect.containsPoint(topRight)) {
          topRightIntersecting = true;
          numCornersIntersecting += 1;
        }
        let bottomLeft = new go.Point(l2, b2);
        if (labelRect.containsPoint(bottomLeft)) {
          bottomLeftIntersecting = true;
          numCornersIntersecting += 1;
        }
        let bottomRight = new go.Point(r2, b2);
        if (labelRect.containsPoint(bottomRight)) {
          bottomRightIntersecting = true;
          numCornersIntersecting += 1;
        }
        // Overlap vector are determine from the number of corners intersecting:
        //  - 0: Edge to edge vector
        //  - 1: Corner to corner vector
        //  - 2: Edge to edge vector
        //  - 4: Corner to corner vector
        if (numCornersIntersecting === 0) {
          //  ---------------
          //  | GraphObject |
          //  |  ---------  |
          //  |  |     |  |
          //  ---------------
          //   | Label |
          //   ---------
          //
          // Downwards vector (side to side)
          //
          // Since no corners of the label/node intersect the moving label,
          // the corners of the moving label intersecting must be found in
          // order to calculate an accurate vector
          // Therefore, the reverse intersection points must be found
          graphObjRect = graphObj.getDocumentBounds();
          let reverseTopLeftIntersecting = false;
          let reverseTopRightIntersecting = false;
          let reverseBottomLeftIntersecting = false;
          let reverseBottomRightIntersecting = false;
          topLeft = new go.Point(l1, t1);
          if (graphObjRect.containsPoint(topLeft)) reverseTopLeftIntersecting = true;
          topRight = new go.Point(r1, t1);
          if (graphObjRect.containsPoint(topRight)) reverseTopRightIntersecting = true;
          bottomLeft = new go.Point(l1, b1);
          if (graphObjRect.containsPoint(bottomLeft)) reverseBottomLeftIntersecting = true;
          bottomRight = new go.Point(r1, b1);
          if (graphObjRect.containsPoint(bottomRight)) reverseBottomRightIntersecting = true;
          if (reverseTopLeftIntersecting && reverseTopRightIntersecting) {
            const upX = 0;
            const upY = b2 - t1;
            if (this._length(upX, upY) > finalMagnitude) finalMagnitude = this._length(upX, upY);
            finalVectorX += upX;
            finalVectorY += upY;
          } else if (reverseTopLeftIntersecting && reverseBottomLeftIntersecting) {
            const leftX = r2 - l1;
            const leftY = 0;
            if (this._length(leftX, leftY) > finalMagnitude) {
              finalMagnitude = this._length(leftX, leftY);
            }
            finalVectorX += leftX;
            finalVectorY += leftY;
          } else if (reverseBottomLeftIntersecting && reverseBottomRightIntersecting) {
            const downX = 0;
            const downY = t2 - b1;
            if (this._length(downX, downY) > finalMagnitude) {
              finalMagnitude = this._length(downX, downY);
            }
            finalVectorX += downX;
            finalVectorY += downY;
          } else if (reverseTopRightIntersecting && reverseBottomRightIntersecting) {
            const rightX = l2 - r1;
            const rightY = 0;
            if (this._length(rightX, rightY) > finalMagnitude) {
              finalMagnitude = this._length(rightX, rightY);
            }
            finalVectorX += rightX;
            finalVectorY += rightY;
          }
          //  ---------------
          //  | GraphObject |
          //  |     ---------
          //  |     |   |   |
          //  ---------------   |
          //      | Label |
          //      ---------
          //
          // Downwards-left vector (corner to corner)
        } else if (numCornersIntersecting === 1) {
          if (topLeftIntersecting) {
            const up_leftX = l2 - r1;
            const up_leftY = t2 - b1;
            if (this._length(up_leftX, up_leftY) > finalMagnitude) {
              finalMagnitude = this._length(up_leftX, up_leftY);
            }
            finalVectorX += up_leftX;
            finalVectorY += up_leftY;
          } else if (topRightIntersecting) {
            const up_rightX = r2 - l1;
            const up_rightY = t2 - b1;
            if (this._length(up_rightX, up_rightY) > finalMagnitude) {
              finalMagnitude = this._length(up_rightX, up_rightY);
            }
            finalVectorX += up_rightX;
            finalVectorY += up_rightY;
          } else if (bottomLeftIntersecting) {
            const down_leftX = l2 - r1;
            const down_leftY = b2 - t1;
            if (this._length(down_leftX, down_leftY) > finalMagnitude) {
              finalMagnitude = this._length(down_leftX, down_leftY);
            }
            finalVectorX += down_leftX;
            finalVectorY += down_leftY;
          } else if (bottomRightIntersecting) {
            const down_rightX = r2 - l1;
            const down_rightY = b2 - t1;
            if (this._length(down_rightX, down_rightY) > finalMagnitude) {
              finalMagnitude = this._length(down_rightX, down_rightY);
            }
            finalVectorX += down_rightX;
            finalVectorY += down_rightY;
          }
          //  ---------------
          //  |  Label  |
          //  |  ---------  |
          //  |  |     |  |
          //  ---------------
          //   |     | <--- GraphObject
          //   ---------
          //
          // Upwards vector (side to side)
        } else if (numCornersIntersecting === 2) {
          if (topLeftIntersecting && topRightIntersecting) {
            const upX = 0;
            const upY = t2 - b1;
            if (this._length(upX, upY) > finalMagnitude) finalMagnitude = this._length(upX, upY);
            finalVectorX += upX;
            finalVectorY += upY;
          } else if (topLeftIntersecting && bottomLeftIntersecting) {
            const leftX = l2 - r1;
            const leftY = 0;
            if (this._length(leftX, leftY) > finalMagnitude) {
              finalMagnitude = this._length(leftX, leftY);
            }
            finalVectorX += leftX;
            finalVectorY += leftY;
          } else if (bottomLeftIntersecting && bottomRightIntersecting) {
            const downX = 0;
            const downY = b2 - t1;
            if (this._length(downX, downY) > finalMagnitude) {
              finalMagnitude = this._length(downX, downY);
            }
            finalVectorX += downX;
            finalVectorY += downY;
          } else if (topRightIntersecting && bottomRightIntersecting) {
            const rightX = r2 - l1;
            const rightY = 0;
            if (this._length(rightX, rightY) > finalMagnitude) {
              finalMagnitude = this._length(rightX, rightY);
            }
            finalVectorX += rightX;
            finalVectorY += rightY;
          }
          //  ---------------
          //  |  Label  |
          //  |  ---------  |
          //  |  |     |  |
          //  |  |     |<-|--- GraphObject
          //  |  |     |  |
          //  |  ---------  |
          //  |       |
          //  ---------------
          //
          // Upwards vector (identical centre points)
        } else if (numCornersIntersecting === 4) {
          if (labelCenter.x - graphObjCenter.x > 0 && labelCenter.y - graphObjCenter.y > 0) {
            const down_rightX = r2 - l1;
            const down_rightY = b2 - t1;
            if (this._length(down_rightX, down_rightY) > finalMagnitude) {
              finalMagnitude = this._length(down_rightX, down_rightY);
            }
            finalVectorX += down_rightX;
            finalVectorY += down_rightY;
          } else if (labelCenter.x - graphObjCenter.x < 0 && labelCenter.y - graphObjCenter.y > 0) {
            const down_leftX = l2 - r1;
            const down_leftY = b2 - t1;
            if (this._length(down_leftX, down_leftY) > finalMagnitude) {
              finalMagnitude = this._length(down_leftX, down_leftY);
            }
            finalVectorX += down_leftX;
            finalVectorY += down_leftY;
          } else if (labelCenter.x - graphObjCenter.x < 0 && labelCenter.y - graphObjCenter.y < 0) {
            const up_leftX = l2 - r1;
            const up_leftY = t2 - b1;
            if (this._length(up_leftX, up_leftY) > finalMagnitude) {
              finalMagnitude = this._length(up_leftX, up_leftY);
            }
            finalVectorX += up_leftX;
            finalVectorY += up_leftY;
          } else if (labelCenter.x - graphObjCenter.x > 0 && labelCenter.y - graphObjCenter.y < 0) {
            const up_rightX = r2 - l1;
            const up_rightY = t2 - b1;
            if (this._length(up_rightX, up_rightX) > finalMagnitude) {
              finalMagnitude = this._length(up_rightX, up_rightY);
            }
            finalVectorX += up_rightX;
            finalVectorY += up_rightY;
          } else if (
            labelCenter.x - graphObjCenter.x === 0 &&
            labelCenter.y - graphObjCenter.y > 0
          ) {
            const downX = 0;
            const downY = b2 - t1;
            if (this._length(downX, downY) > finalMagnitude) {
              finalMagnitude = this._length(downX, downY);
            }
            finalVectorX += downX;
            finalVectorY += downY;
          } else if (
            labelCenter.x - graphObjCenter.x === 0 &&
            labelCenter.y - graphObjCenter.y < 0
          ) {
            const upX = 0;
            const upY = t2 - b1;
            if (this._length(upX, upY) > finalMagnitude) finalMagnitude = this._length(upX, upY);
            finalVectorX += upX;
            finalVectorY += upY;
          } else if (
            labelCenter.x - graphObjCenter.x > 0 &&
            labelCenter.y - graphObjCenter.y === 0
          ) {
            const rightX = r2 - l1;
            const rightY = 0;
            if (this._length(rightX, rightY) > finalMagnitude) {
              finalMagnitude = this._length(rightX, rightY);
            }
            finalVectorX += rightX;
            finalVectorY += rightY;
          } else if (
            labelCenter.x - graphObjCenter.x < 0 &&
            labelCenter.y - graphObjCenter.y === 0
          ) {
            const leftX = l2 - r1;
            const leftY = 0;
            if (this._length(leftX, leftY) > finalMagnitude) {
              finalMagnitude = this._length(leftX, leftY);
            }
            finalVectorX += leftX;
            finalVectorY += leftY;
          } else if (
            labelCenter.x - graphObjCenter.x === 0 &&
            labelCenter.y - graphObjCenter.y === 0
          ) {
            // If two labels are exactly on top of each other --> just move up
            const upX = 0;
            const upY = t2 - b1;
            if (this._length(upX, upY) > finalMagnitude) finalMagnitude = this._length(upX, upY);
            finalVectorX += upX;
            finalVectorY += upY;
          }
        }
      }
      //  -------
      //  |   |<--- Label
      //  -------
      //   -------
      //   |   |<--- GraphObject
      //   -------
      //
      // Upwards-left vector (centre to centre)
      // If the nearby object doesn't intersect, a smaller vector is calculated
      else {
        const distance = labelCenter.distanceSquaredPoint(graphObjCenter);
        let vectorX = labelCenter.x - graphObjCenter.x;
        let vectorY = labelCenter.y - graphObjCenter.y;
        const tempVectorX = vectorX;
        const tempVectorY = vectorY;
        vectorX = this._normalize(tempVectorX, tempVectorX, tempVectorY);
        vectorY = this._normalize(tempVectorY, tempVectorX, tempVectorY);
        const size = this._sizeOfGraphObj(graphObj);
        vectorX = this._scaleBy(vectorX, (19 * size) / distance);
        vectorY = this._scaleBy(vectorY, (19 * size) / distance);
        finalVectorX += vectorX;
        finalVectorY += vectorY;
        finalMagnitude = this._length(finalVectorX, finalVectorY);
      }
    });
    //  |
    //  |   ---------
    //  |   | Label |
    //  |   ---------
    //  |
    //  | <--- Link
    //  |
    //
    //  Leftwards vector (centre to orthogonal point)
    // Add vector with magnitude and direction towards corresponding link
    if (this.precision === 5) {
      const link = label.panel;
      if (!(link instanceof go.Link)) return [];
      if (link.path === null) return [];
      // Calculate centre point of label
      const pointLabel = link.path.getLocalPoint(label.getDocumentPoint(go.Spot.Center));
      // Calculate fraction (0-1) down link for label
      const fractionLinkForLabel = link.geometry.getFractionForPoint(pointLabel);
      // Caluclate point on link for that fraction
      const pointLink = link.geometry.getPointAlongPath(fractionLinkForLabel);
      const towardsLinkX = pointLink.x - pointLabel.x;
      const towardsLinkY = pointLink.y - pointLabel.y;
      if (this._length(towardsLinkX, towardsLinkY) > finalMagnitude) {
        finalMagnitude = this._length(towardsLinkX, towardsLinkY);
      }
      finalVectorX += towardsLinkX;
      finalVectorY += towardsLinkY;
    }
    // After the summation of all vectors
    // if the sum creates a zero vector, check nearby positions
    if (finalVectorX === 0 || finalVectorY === 0) {
      this._checkMoves(label, 20);
      return [finalVectorX, finalVectorY];
    }
    // Normalize and scale the vector to an appropriate size
    const tVectorX = finalVectorX;
    const tVectorY = finalVectorY;
    finalVectorX = this._normalize(tVectorX, tVectorX, tVectorY);
    finalVectorY = this._normalize(tVectorY, tVectorX, tVectorY);
    if (this._nearby.size === 1) {
      finalVectorX = this._scaleBy(finalVectorX, finalMagnitude);
      finalVectorY = this._scaleBy(finalVectorY, finalMagnitude);
    } else {
      finalVectorX = this._scaleBy(finalVectorX, finalMagnitude / 2);
      finalVectorY = this._scaleBy(finalVectorY, finalMagnitude / 2);
    }
    return [finalVectorX, finalVectorY];
  }

  /**
   * Recursive vector process towards local minima
   *
   * @hidden @internal
   * @param {GraphObject} label
   * @returns {void}
   */
  private _differentialCalculatorVector(label: go.GraphObject) {
    if (this._checkVectorMove(label)) this._differentialCalculatorVector(label);
    else {
      return;
    }
  }

  /**
   * Check to see if the precision vector movement results
   * in a better position score
   *
   * @hidden @internal
   * @param {GraphObject} label
   * @returns {boolean}
   */
  private _checkVectorMove(label: go.GraphObject) {
    const link = label.panel;
    if (!(link instanceof go.Link)) return false;
    if (link.path === null) return false;
    // Calculate centre point of label
    const pointLabel = link.path.getLocalPoint(label.getDocumentPoint(go.Spot.Center));
    // Calculate fraction (0-1) down link for label
    const fractionLinkForLabel = link.geometry.getFractionForPoint(pointLabel);
    // Caluclate point on link for that fraction
    const pointLink = link.geometry.getPointAlongPath(fractionLinkForLabel);
    const size = this._sizeOfGraphObj(label);
    const distance = Math.pow(pointLabel.distanceSquaredPoint(pointLink), 0.5);
    // Gradient vector for distance_to_link (always points towards link)
    let gradVectorX =
      Math.pow(2, (4 * size + 10 * distance) / (size + 1)) * (pointLink.x - pointLabel.x);
    let gradVectorY =
      Math.pow(2, (4 * size + 10 * distance) / (size + 1)) * (pointLink.y - pointLabel.y);
    gradVectorX = this._scaleBy(gradVectorX, (0.01 * 10 * Math.log(2)) / (size * distance + 1));
    gradVectorY = this._scaleBy(gradVectorY, (0.01 * 10 * Math.log(2)) / (size * distance + 1));
    const overlapVector = this._vectorMovement(label);
    let vectorX = gradVectorX + overlapVector[0];
    let vectorY = gradVectorY + overlapVector[1];
    if (vectorX === 0 && vectorY === 0) return false;
    const tempVectorX = vectorX;
    const tempVectorY = vectorY;
    vectorX = this._normalize(tempVectorX, tempVectorX, tempVectorY);
    vectorY = this._normalize(tempVectorY, tempVectorX, tempVectorY);
    vectorX = this._scaleBy(vectorX, this.precision);
    vectorY = this._scaleBy(vectorY, this.precision);
    const currentScore = this._cost(label);
    this._changeFocusLabel(label, -vectorX, -vectorY);
    const movedScore = this._cost(label);
    if (movedScore < currentScore) return true;
    else {
      this._changeFocusLabel(label, vectorX, vectorY);
      return false;
    }
  }

  /**
   * Check along the eight main directions and starting point
   * for the optimum label position
   *
   * @hidden @internal
   * @param {GraphObject} label
   * @param {number} move
   * @returns {boolean}
   */
  private _checkMoves(label: go.GraphObject, move: number) {
    this._changeFocusLabel(label, 0, move); // if moved up
    const up = this._cost(label);
    this._changeFocusLabel(label, 0, -2 * move); // if moved down
    const down = this._cost(label);
    this._changeFocusLabel(label, -move, move); // if moved left
    const left = this._cost(label);
    this._changeFocusLabel(label, 2 * move, 0); // if moved right
    const right = this._cost(label);
    this._changeFocusLabel(label, -(move / Math.SQRT2 + move), move / Math.SQRT2); // if moved up_left
    const up_left = this._cost(label);
    this._changeFocusLabel(label, (2 * move) / Math.SQRT2, 0); // if moved up_right
    const up_right = this._cost(label);
    this._changeFocusLabel(label, 0, (-2 * move) / Math.SQRT2); // if moved down_right
    const down_right = this._cost(label);
    this._changeFocusLabel(label, (-2 * move) / Math.SQRT2, 0); // if moved down_left
    const down_left = this._cost(label);
    // Return to original position
    this._changeFocusLabel(label, move / Math.SQRT2, move / Math.SQRT2);
    const original = this._cost(label);
    const minCOST = Math.min(
      up,
      down,
      left,
      right,
      original,
      up_left,
      up_right,
      down_left,
      down_right
    );
    // Return false if the best position is the original position
    if (minCOST === original) return false;
    else if (minCOST === up) {
      this._changeFocusLabel(label, 0, move);
      return true;
    } else if (minCOST === down) {
      this._changeFocusLabel(label, 0, -move);
      return true;
    } else if (minCOST === left) {
      this._changeFocusLabel(label, -move, 0);
      return true;
    } else if (minCOST === right) {
      this._changeFocusLabel(label, move, 0);
      return true;
    } else if (minCOST === up_left) {
      this._changeFocusLabel(label, -move / Math.SQRT2, move / Math.SQRT2);
      return true;
    } else if (minCOST === up_right) {
      this._changeFocusLabel(label, move / Math.SQRT2, move / Math.SQRT2);
      return true;
    } else if (minCOST === down_left) {
      this._changeFocusLabel(label, -move / Math.SQRT2, -move / Math.SQRT2);
      return true;
    } else if (minCOST === down_right) {
      this._changeFocusLabel(label, move / Math.SQRT2, -move / Math.SQRT2);
      return true;
    }
    return false;
  }

  /**
   * Finds Group for cases when detectGroups is true, but when the Group
   * should still be ignored
   *
   * @hidden @internal
   * @param {GraphObject} label
   * @returns
   */
  private _commonGroup(label: go.GraphObject) {
    if (this.detectGroups === true) {
      const link = label.panel;
      if (label.panel === null) return null;
      if (link instanceof go.Link) {
        const toNode = link.toNode;
        if (toNode === null) return null;
        const fromNode = link.fromNode;
        if (fromNode === null) return null;
        const group = toNode.findCommonContainingGroup(fromNode);
        return group;
      }
    } else {
      return null;
    }
    return null;
  }

  /**
   * Calulate the half-diagonal distance of a label
   *
   * @hidden @internal
   * @param {GraphObject} label
   * @returns {number}
   */
  private _sizeOfGraphObj(label: go.GraphObject) {
    const labelRect = label.getDocumentBounds();
    // calculate the diagonal distance of the label
    const labelSize = Math.hypot(labelRect.width / 2, labelRect.height / 2);
    return labelSize;
  }

  /**
   * Change the alignmentFocus value of a label
   *
   * @hidden @internal
   * @param {GraphObject} label
   * @param {number} x
   * @param {number} y
   * @returns {void}
   */
  private _changeFocusLabel(label: go.GraphObject, x: number, y: number) {
    let _a;
    const offX = label.alignmentFocus.offsetX;
    const offY = label.alignmentFocus.offsetY;
    label.alignmentFocus = new go.Spot(0.5, 0.5, offX + x, offY + y);
    (_a = label.part) === null || _a === void 0 ? void 0 : _a.ensureBounds();
    return;
  }

  /**
   * Finds out whether two Rect intersect or not
   *
   * @hidden @internal
   * @param {Rect} firstRect
   * @param {Rect} secondRect
   * @returns {boolean}
   */
  private _doIntersect(
    l1: number,
    t1: number,
    r1: number,
    b1: number,
    l2: number,
    t2: number,
    r2: number,
    b2: number
  ) {
    if (l1 > r2 || l2 > r1) return false;
    if (t1 > b2 || t2 > b1) return false;
    return true;
  }

  /**
   * Scale a vector
   *
   * @hidden @internal
   * @param {number} component
   * @param {number} scale
   * @returns {number}
   */
  private _scaleBy(component: number, scale: number) {
    return component * scale;
  }

  /**
   * Find the magnitude of a vector
   *
   * @hidden @internal
   * @param {number} x
   * @param {number} y
   * @returns {number}
   */
  private _length(x: number, y: number) {
    return Math.pow(Math.pow(x, 2) + Math.pow(y, 2), 0.5);
  }

  /**
   * Normalize a vector
   *
   * @hidden @internal
   * @param {number} component
   * @param {number} x
   * @param {number} y
   * @returns {number}
   */
  private _normalize(component: number, x: number, y: number) {
    return this._scaleBy(component, 1 / this._length(x, y)); // Zero vector exception checked in main code
  }
}
